import os
import tempfile
import logging
import sys
import time
import shutil
import zipfile
from urllib.request import urlopen
from dynamic_preferences.registries import global_preferences_registry as gpr
from simo.core.models import Gateway
from .gateways import ZwaveGatewayHandler



def get_latest_ozw_library():
    stopped_gateways = []
    for gateway in Gateway.objects.filter(type=ZwaveGatewayHandler.uid):
        if gateway.status == 'running':
            gateway.stop()
            stopped_gateways.append(gateway)

    if stopped_gateways:
        if len(stopped_gateways) == 1:
            print("Stopping running gateway")
        else:
            print("Stopping running gateways")
        # give some time for gateways to stop
        time.sleep(2)

    configs_path = gpr.manager()['zwave__ozwave_config_path_community']

    dest = tempfile.mkdtemp()
    dest_file = os.path.join(dest, 'open-zwave.zip')
    print("Getting latest configs from OpenZwave Github project")
    try:
        req = urlopen(
            'https://codeload.github.com/OpenZWave/open-zwave/zip/master')
        with open(dest_file, 'wb') as f:
            f.write(req.read())
        zip_ref = zipfile.ZipFile(dest_file, 'r')
        zip_ref.extractall(dest)
        zip_ref.close()
    except Exception:
        print("Can't get zip from github. Cancelling.")
        try:
            shutil.rmtree(dest)
        except Exception:
            pass
        return "Can't get zip from github. Cancelling."

    if os.path.isdir(configs_path):
        # Try to remove old config
        try:
            shutil.rmtree(configs_path)
        except Exception:
            print("Can't remove old config directory")
            return "Can't remove old config directory"

    try:
        shutil.copytree(os.path.join(dest, 'open-zwave-master', 'config'),
                        configs_path)
    except Exception:
        print("Can't copy to %s", configs_path)
        return "Can't copy to %s", configs_path

    try:
        with open(os.path.join(configs_path,
                               'pyozw_config.version'), 'w') as f:
            f.write(time.strftime("%Y-%m-%d %H:%M"))
    except Exception:
        msg = "Can't update %s" % os.path.join(
            configs_path, 'pyozw_config.version'
        )
        print(msg)
        return msg

    try:
        with open(os.path.join(configs_path, '__init__.py'),
                  'a') as f:
            f.write(
                "#This file is part of **python-openzwave** project https://github.com/OpenZWave/python-openzwave.")
    except Exception:
        msg = "Can't create %s" % os.path.join(configs_path, '__init__.py')
        return msg

    shutil.rmtree(dest)

    if stopped_gateways:
        if len(stopped_gateways) == 1:
            print("Starting gateway")
        else:
            print("Starting gateways")
        for gateway in stopped_gateways:
            gateway.start()



import sys
from django.db import models
from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver
from simo.core.models import Gateway, Component
from simo.core.events import GatewayObjectCommand




class ZwaveNode(models.Model):
    name = models.CharField(
        max_length=200, blank=True, default=''
    )
    gateway = models.ForeignKey(
        Gateway, on_delete=models.CASCADE, related_name='zwave_nodes'
    )
    node_id = models.PositiveIntegerField(
        primary_key=True, unique=True, editable=False
    )
    product_name = models.CharField(max_length=200, editable=False)
    product_type = models.CharField(max_length=200, editable=False)
    battery_level = models.PositiveIntegerField(null=True, editable=False)
    alive = models.BooleanField(default=True)
    stats = models.JSONField(default=dict)
    is_new = models.BooleanField(default=True, editable=False)

    def __str__(self):
        return '[%d] %s' % (
            self.node_id, self.name if self.name else self.product_name
        )

    def kill_node(self):
        print("kill node %s" % str(self.node_id))
        GatewayObjectCommand(
            self.gateway, zwave_command='remove_failed_node', node_id=self.node_id
        ).publish()

    def request_network_update(self):
        GatewayObjectCommand(
            self.gateway, zwave_command='request_network_update', node_id=self.node_id
        ).publish()

    def send_node_information(self):
        GatewayObjectCommand(
            self.gateway, zwave_command='send_node_information', node_id=self.node_id
        ).publish()

    def has_node_failed(self):
        GatewayObjectCommand(
            self.gateway, zwave_command='has_node_failed', node_id=self.node_id
        ).publish()


class NodeValue(models.Model):
    node = models.ForeignKey(
        ZwaveNode, on_delete=models.CASCADE, related_name='node_values'
    )
    genre = models.CharField(max_length=100, db_index=True, null=True)
    value_id = models.BigIntegerField()
    index = models.PositiveIntegerField(
        null=True, blank=True, editable=False, db_index=True
    )
    label = models.CharField(max_length=200)
    is_read_only = models.BooleanField()
    type = models.CharField(max_length=100)
    units = models.CharField(max_length=100, blank=True)
    value_choices = models.JSONField(default=list, editable=False)
    value = models.JSONField(null=True, blank=True)
    value_new = models.JSONField(null=True, blank=True)
    name = models.CharField(
        max_length=100, null=True, blank=True,
        help_text="Give it a name to allow use in components."
    )
    component = models.ForeignKey(
        Component, on_delete=models.SET_NULL,
        null=True, blank=True, related_name='zwave_node_val'
    )

    class Meta:
        unique_together = 'node', 'value_id'
        ordering = '-genre', 'index',

    def __str__(self):
        return '%s | [%d] %s' % (
            str(self.node), self.id, self.name if self.name else self.label
        )

@receiver(post_save, sender=NodeValue)
def set_component_value_of_node_value(sender, instance, created, *args, **kwargs):
    '''Set component value to zwave node value on component creation'''
    if created:
        return
    org = NodeValue.objects.get(pk=instance.pk)
    if instance.component and org.component != instance.component:
        instance.component.set(instance.value)



@receiver(post_delete, sender=ZwaveNode)
def node_post_delete(sender, instance, *args, **kwargs):
    GatewayObjectCommand(
        instance.gateway, zwave_command='remove_failed_node', node_id=instance.node_id
    ).publish()


@receiver(post_save, sender=Gateway)
def update_zwave_library_on_new_zwave_gateway(
    sender, instance, created, *args, **kwargs
):
    if not created:
        return
    from .gateways import ZwaveGatewayHandler
    if instance.type == ZwaveGatewayHandler.uid and 'test' not in sys.argv:
        from .utils import get_latest_ozw_library
        get_latest_ozw_library()




# Generated by Django 3.2.9 on 2022-07-07 14:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('simo_zwave', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='nodevalue',
            name='genre',
            field=models.CharField(db_index=True, max_length=100, null=True),
        ),
    ]



from simo.core.controllers import (
    BinarySensor, NumericSensor,
    Switch, Dimmer, RGBWLight
)
from .gateways import ZwaveGatewayHandler
from .forms import (
    BasicZwaveComponentConfigForm, ZwaveKnobComponentConfigForm,
    RGBLightComponentConfigForm, ZwaveNumericSensorConfigForm,
    ZwaveSwitchConfigForm
)

class ZwaveBinarySensor(BinarySensor):
    gateway_class = ZwaveGatewayHandler
    config_form = BasicZwaveComponentConfigForm

    def _receive_from_device(self, val):
        return super()._receive_from_device(bool(val))

class ZwaveNumericSensor(NumericSensor):
    gateway_class = ZwaveGatewayHandler
    config_form = ZwaveNumericSensorConfigForm


class ZwaveSwitch(Switch):
    gateway_class = ZwaveGatewayHandler
    config_form = ZwaveSwitchConfigForm

    def _receive_from_device(self, val):
        return super()._receive_from_device(bool(val))


class ZwaveDimmer(Dimmer):
    gateway_class = ZwaveGatewayHandler
    config_form = ZwaveKnobComponentConfigForm

    def _send_to_device(self, value):
        conf = self.component.config

        com_amplitude = conf.get('max', 1.0) - conf.get('min', 0.0)
        float_value = (value - conf.get('min', 0.0)) / com_amplitude

        zwave_amplitude = conf.get('zwave_max', 99.0) - conf.get('zwave_min', 0.0)
        set_val = float_value * zwave_amplitude + conf.get('zwave_min', 0.0)

        return super()._send_to_device(set_val)

    def _receive_from_device(self, val):
        conf = self.component.config

        zwave_amplitude = conf.get('zwave_max', 99.0) - conf.get('zwave_min', 0.0)
        float_value = (val - conf.get('zwave_min', 0.0)) / zwave_amplitude

        com_amplitude = conf.get('max', 99.0) - conf.get('min', 0.0)
        set_val = float_value * com_amplitude + conf.get('min', 0.0)

        return super()._receive_from_device(set_val)


class ZwaveRGBWLight(RGBWLight):
    gateway_class = ZwaveGatewayHandler
    config_form = RGBLightComponentConfigForm

    def _receive_from_device(self, val):
        # TODO: need to addapt to map type RGBWLight value.
        return super()._receive_from_device(val)



from django.contrib.auth.decorators import login_required
from django.http import HttpResponseForbidden, HttpResponse
from django.shortcuts import redirect
from django.contrib import messages
from .utils import get_latest_ozw_library


@login_required
def update_zwave_library(request):
    if not request.user.is_superuser:
        return HttpResponseForbidden()
    try:
        error = get_latest_ozw_library()
    except Exception as e:
        messages.error(request, str(e))
    else:
        if error:
            messages.error(request, str(error))
        else:
            messages.success(request, "Open OZW devices library updated!")
    if request.META.get('HTTP_REFERER'):
        return redirect(request.META.get('HTTP_REFERER'))
    return HttpResponse('Updated!')




# Generated by Django 2.2.12 on 2021-09-24 11:57

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('core', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ZwaveNode',
            fields=[
                ('name', models.CharField(blank=True, default='', max_length=200)),
                ('node_id', models.PositiveIntegerField(editable=False, primary_key=True, serialize=False, unique=True)),
                ('product_name', models.CharField(editable=False, max_length=200)),
                ('product_type', models.CharField(editable=False, max_length=200)),
                ('battery_level', models.PositiveIntegerField(editable=False, null=True)),
                ('alive', models.BooleanField(default=True)),
                ('stats', models.JSONField(default=dict)),
                ('is_new', models.BooleanField(default=True, editable=False)),
                ('gateway', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='zwave_nodes', to='core.Gateway')),
            ],
        ),
        migrations.CreateModel(
            name='NodeValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('genre', models.CharField(db_index=True, max_length=100)),
                ('value_id', models.BigIntegerField()),
                ('index', models.PositiveIntegerField(blank=True, db_index=True, editable=False, null=True)),
                ('label', models.CharField(max_length=200)),
                ('is_read_only', models.BooleanField()),
                ('type', models.CharField(max_length=100)),
                ('units', models.CharField(blank=True, max_length=100)),
                ('value_choices', models.JSONField(default=list, editable=False)),
                ('value', models.JSONField(blank=True, null=True)),
                ('value_new', models.JSONField(blank=True, null=True)),
                ('name', models.CharField(blank=True, help_text='Give it a name to allow use in components.', max_length=100, null=True)),
                ('component', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='zwave_node_val', to='core.Component')),
                ('node', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='node_values', to='simo_zwave.ZwaveNode')),
            ],
            options={
                'ordering': ('-genre', 'index'),
                'unique_together': {('node', 'value_id')},
            },
        ),
    ]



from django import forms


class AdminNodeValueValueWidget(forms.TextInput):

    def __init__(self, value_new, *args, **kwargs):
        self.value_new = value_new
        super().__init__(*args, **kwargs)

    def render(self, name, value, attrs=None, renderer=None):
        org = super().render(name, value, attrs, renderer)
        if self.value_new != value:
            return org + '<br>&nbsp;&nbsp;<strong>%s</strong> - pending' % self.value_new
        return org


class AdminNodeValueSelectWidget(forms.Select):

    def __init__(self, value_new, *args, **kwargs):
        self.value_new = value_new
        super().__init__(*args, **kwargs)

    def render(self, name, value, attrs=None, renderer=None):
        org = super().render(name, value, attrs, renderer)
        if self.value_new != value:
            return org + '<br>&nbsp;&nbsp;<strong>%s</strong> - pending' % self.value_new
        return org


import os
import logging
import json
import sys
from openzwave.network import ZWaveNetwork
from openzwave.option import ZWaveOption
import paho.mqtt.client as mqtt
from pydispatch import dispatcher
import time
from django.conf import settings
from simo.core.models import Component
from simo.core.gateways import BaseGatewayHandler
from simo.core.events import GatewayObjectCommand, get_event_obj
from simo.conf import dynamic_settings
from .forms import ZwaveGatewayForm
from .models import ZwaveNode, NodeValue


class ZwaveGatewayHandler(BaseGatewayHandler):
    name = "Zwave"
    config_form = ZwaveGatewayForm
    auto_create = True
    network = None

    def run(self, exit):
        print("Starting Zwave Gateway")
        mqtt_client = mqtt.Client()
        mqtt_client.username_pw_set('root', settings.SECRET_KEY)

        try:
            config_path = dynamic_settings['zwave__ozwave_config_path_community']
            if not os.path.exists(config_path):
                os.makedirs(config_path)
            user_path = dynamic_settings['zwave__ozwave_config_path_user']
            if not os.path.exists(user_path):
                os.makedirs(user_path)

            options = ZWaveOption(
                device=self.gateway_instance.config['device'],
                config_path=dynamic_settings['zwave__ozwave_config_path_community'],
                user_path=dynamic_settings['zwave__ozwave_config_path_user']
            )
            options.addOptionString(
                'NetworkKey', dynamic_settings['zwave__netkey'], False
            )
            options.lock()

            self.network = ZWaveNetwork(options, autostart=False)
            dispatcher.connect(
                self.network_ready, ZWaveNetwork.SIGNAL_NETWORK_AWAKED
            )
            self.network.start()


            mqtt_client.on_connect = self.on_mqtt_connect
            mqtt_client.on_message = self.on_mqtt_message
            mqtt_client.connect(
                host=settings.MQTT_HOST, port=settings.MQTT_PORT
            )
            mqtt_client.loop_start()

            seconds_counter = 0
            while not exit.is_set():
                time.sleep(1)
                if seconds_counter > 30:
                    self.perform_periodic_maintenance()
                    seconds_counter = 0
                else:
                    seconds_counter += 1

            print("I must stop now!")
            mqtt_client.loop_stop()
            print("MQTT stopped!")
            self.network.stop()
            print("Network stopped!")

        except SystemExit:
            print("STOPPING.....")
            mqtt_client.loop_stop()
            self.network.stop()
            raise SystemExit()

    def on_mqtt_connect(self, mqtt_client, userdata, flags, rc):
        print("Connected to mqtt with result code " + str(rc))
        command = GatewayObjectCommand(self.gateway_instance)
        mqtt_client.subscribe(command.get_topic())

    def on_mqtt_message(self, client, userdata, msg):
        payload = json.loads(msg.payload)

        if 'zwave_command' in payload:
            if not hasattr(self.network, payload['zwave_command']):
                return
            zwave_command = getattr(
                self.network.controller, payload['zwave_command']
            )
            if 'node_id' in payload:
                kwargs = {'node_id': payload['node_id']}
            else:
                kwargs = {}
            try:
                zwave_command(**kwargs)
            except Exception as e:
                print(e, file=sys.stderr)
                return
            else:
                print("Command %s initiated!" % payload['command'])
                self.gateway_instance.refresh_from_db()
                if payload['command'] == 'cancel_command' \
                and 'last_controller_command' in self.gateway_instance.config:
                    self.gateway_instance.config.pop('last_controller_command')
                    self.gateway_instance.save()
            return

        if 'bulk_send' in payload:
            components = {c.id: c for c in Component.objects.filter(
                gateway__type=self.uid,
                id__in=[int(id) for id in payload['bulk_send'].keys()]
            )}
            for comp_id, value in payload['bulk_send'].items():
                if int(comp_id) not in components:
                    continue
                comp = components[int(comp_id)]
                try:
                    value = comp.controller._string_to_vals(str(value))[0]
                except:
                    pass
                self.send_val(components[int(comp_id)], value)
            return

        target = get_event_obj(payload)
        if isinstance(target, Component):
            if 'set_val' not in payload:
                return
            self.send_val(target, payload['set_val'])

    def send_val(self, target, val):
        if isinstance(target, NodeValue):
            node_val = target
        else:
            try:
                node_val = NodeValue.objects.get(
                    pk=target.config.get('zwave_item', 0)
                )
            except:
                return
        try:
            nodev = self.network.nodes[
                node_val.node.node_id
            ].values[node_val.value_id]
            nodev.data = val
        except Exception as e:
            logging.error(e, exc_info=True)
            pass

    def update_node_stats(self, node_model):

        try:
            zwave_node = self.network.nodes[node_model.node_id]
            node_model.stats = zwave_node.stats
        except KeyError:
            node_model.alive = False
        else:
            node_model.product_name = zwave_node.product_name
            node_model.product_type = zwave_node.type
            node_model.alive = not zwave_node.is_failed
            node_model.battery_level = zwave_node.get_battery_level()

        node_model.save()

        for node_val in node_model.node_values.all().exclude(component=None):
            node_val.component.alive = node_model.alive
            node_val.component.battery_level = node_model.battery_level
            node_val.component.save()

    def perform_periodic_maintenance(self):
        self.gateway_instance.refresh_from_db()
        last_controller_command = self.gateway_instance.config.get(
            'last_controller_command'
        )
        if last_controller_command \
        and time.time() - last_controller_command > 20:
            self.network.controller.cancel_command()
            self.gateway_instance.config.pop('last_controller_command')
            self.gateway_instance.save()
            print("CANCELED ALL COMMANDS!!!!")
        for node_model in self.gateway_instance.zwave_nodes.all():
            self.update_node_stats(node_model)


    def update_node_val(self, node, val):
        node = ZwaveNode.objects.filter(node_id=node.node_id).first()
        if not node:
            self.update_node_model(node)

        data = {
            'genre': val.genre, 'type': val.type, 'label': val.label,
            'is_read_only': val.is_read_only, 'index': val.index,
            'units': val.units, 'value': val.data,
            'value_new': val.data,
            'value_choices': list(val.data_items) if val.type == 'List' else [],
        }
        node_val, new = NodeValue.objects.get_or_create(
            node=node, value_id=val.value_id, defaults=data
        )
        if not new:
            for attr, v in data.items():
                setattr(node_val, attr, v)
                node_val.save()
        if node_val.label.lower().startswith('battery lev') \
                and val.units == '%':
            node_val.node.battery_level = val.data
            node_val.node.save()

        if node_val.component:
            try:
                node_val.component.controller._receive_from_device(
                    node_val.value
                )
            except Exception as e:
                logging.exception("Unable to save value to component!")
                logging.error(e, exc_info=True)



    def update_node_model(self, node):
        if node.product_type == '0x0001' or node.node_id in (1, 255):
            # Static PC Controller
            return

        node_model, new = ZwaveNode.objects.get_or_create(
            node_id=node.node_id, gateway=self.gateway_instance, defaults={
                'product_name': node.product_name,
                'product_type': node.type
            }
        )
        self.update_node_stats(node_model)

        for key, val in node.values.items():
            self.update_node_val(node, val)

    def network_ready(self, network):
        dispatcher.connect(self.node_update, ZWaveNetwork.SIGNAL_NODE_ADDED)
        dispatcher.connect(self.node_update, ZWaveNetwork.SIGNAL_NODE)
        dispatcher.connect(self.node_removed, ZWaveNetwork.SIGNAL_NODE_REMOVED)
        dispatcher.connect(self.value_update, ZWaveNetwork.SIGNAL_VALUE)

        for node_id, node in self.network.nodes.items():
            self.update_node_model(node)

    def node_update(self, network, node):
        self.update_node_model(node)

    def node_removed(self, network, node):
        try:
            node_model = ZwaveNode.objects.get(
                node_id=node.node_id, gateway=self.gateway_instance
            )
        except ZwaveNode.DoesNotExist:
            pass
        else:
            node_model.delete()

    def value_update(self, network, node, value):
        self.update_node_val(node, value)





from django import forms
from django.db.models import Q
from dal import forward
from django.utils.translation import gettext_lazy as _
from simo.core.utils.validators import validate_slaves
from simo.core.utils.form_widgets import AdminReadonlyFieldWidget, EmptyFieldWidget
from simo.core.forms import BaseGatewayForm, BaseComponentForm, NumericSensorForm
from simo.core.models import Gateway
from simo.core.events import GatewayObjectCommand
from simo.core.models import Component
from simo.core.form_fields import (
    Select2ModelChoiceField, Select2ListChoiceField,
    Select2ModelMultipleChoiceField
)
from .models import ZwaveNode, NodeValue
from .widgets import AdminNodeValueValueWidget, AdminNodeValueSelectWidget


class ZwaveLibraryUpdateBtnWidget(forms.TextInput):
    template_name = 'admin/zwave/library_update_btn.html'


class ZwaveGatewayForm(BaseGatewayForm):
    device = forms.CharField(initial="/dev/ttyACM0")
    library = forms.Field(required=False, widget=ZwaveLibraryUpdateBtnWidget())


class AdminNodeValueInlineForm(forms.ModelForm):
    value = forms.CharField(required=False)

    class Meta:
        model = NodeValue
        fields = 'label', 'value', 'units', 'genre',

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.org_val = self.instance.value
        if self.instance.is_read_only:
            self.fields['value'].widget = AdminReadonlyFieldWidget()
        else:
            if self.instance.type in ('Button', 'Bool'):
                self.fields['value'].widget = AdminNodeValueSelectWidget(
                    self.instance.value_new,
                    choices=[('True', 'True'), ('False', 'False')]
                )
            elif self.instance.type == 'List' and self.instance.value_choices:
                self.fields['value'].widget = AdminNodeValueSelectWidget(
                    self.instance.value_new,
                    choices=[('%s' % c, c) for c in self.instance.value_choices],
                    attrs={'style': 'width:200px'}
                )
            else:
                self.fields['value'].widget = AdminNodeValueValueWidget(
                    self.instance.value_new
                )


        if self.instance.genre != 'User':
            self.fields['name'].widget = EmptyFieldWidget()


    def clean_value(self):

        val = self.cleaned_data['value']

        if self.instance.type in ('Button', 'Bool'):
            if val == 'True':
                val = True
            elif val == 'False':
                val = False
            else:
                raise forms.ValidationError("Must be True or False")


        elif self.instance.type in ('Byte', 'Int', 'Short'):
            try:
                val = int(val)
            except:
                raise forms.ValidationError("Must be a number")

            if self.instance.type == 'Byte':
                if val < 0 or val > 255:
                    raise forms.ValidationError(
                        "Must be a number between 0 and 255"
                    )

            if self.instance.type == 'Short':
                try:
                    val = int(val)
                except:
                    raise forms.ValidationError(
                        "Must be an integer between -32768 and 32768"
                    )
                else:
                    if int(val) > 32768 or int(val) < -32768:
                        raise forms.ValidationError(
                            "Must be an integer between -32768 and 32768"
                        )

        elif self.instance.type == 'Decimal':
            try:
                val = float(val)
            except:
                raise forms.ValidationError("Must be a decimal number")

        return val

    def save(self, commit=True):
        self.instance.value = self.org_val
        set_val = False
        if not self.instance.is_read_only \
        and self.cleaned_data['value'] != self.org_val:
            self.instance.value_new = self.cleaned_data['value']
            set_val = True
        obj = super().save(commit=commit)
        if set_val:
            for gateway in Gateway.objects.filter(type__startswith='simo_zwave'):
                GatewayObjectCommand(
                    gateway, self.instance,
                    set_val=self.cleaned_data['value']
                ).publish()
        return obj


class ZwaveGatewaySelectForm(forms.Form):
    gateway = forms.ModelChoiceField(
        queryset=Gateway.objects.filter(type__startswith='simo_zwave')
    )



class BasicZwaveComponentConfigForm(BaseComponentForm):
    zwave_item = forms.ModelChoiceField(queryset=NodeValue.objects.all())

    node_value = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        qs = NodeValue.objects.filter(
            node__gateway=self.gateway, genre='User', name__isnull=False,
        ).filter(
            Q(component=None) | Q(component=self.instance)
        ).order_by('node_id', 'id')
        self.fields['zwave_item'].queryset = qs


    def clean_zwave_val(self):
        if self.cleaned_data['zwave_item'].component in (
            None, self.instance
        ):
            return self.cleaned_data['zwave_item']

        raise forms.ValidationError(
            'Zwave val already has other component assigned to it'
        )

    def save(self, commit=True):
        self.instance.value = self.cleaned_data['zwave_item'].value
        org_zwave_item = None
        if not self.instance.pk:
            self.instance.value_units = self.cleaned_data['zwave_item'].units
        else:
            org_zwave_item = NodeValue.objects.get(
                pk=self.instance.config.get('zwave_item')
            )
        obj = super().save(commit=commit)
        self.cleaned_data['zwave_item'].component = obj
        self.cleaned_data['zwave_item'].save()
        if org_zwave_item \
        and org_zwave_item.pk != self.cleaned_data['zwave_item'].pk:
            org_zwave_item.component = None
            org_zwave_item.save()
        return obj


class ZwaveNumericSensorConfigForm(BasicZwaveComponentConfigForm, NumericSensorForm):
    pass


class ZwaveSwitchConfigForm(BasicZwaveComponentConfigForm):
    slaves = Select2ModelMultipleChoiceField(
        queryset=Component.objects.filter(
            base_type__in=(
                'dimmer', 'switch', 'blinds', 'script'
            )
        ),
        url='autocomplete-component',
        forward=[
            forward.Const(['dimmer', 'switch', 'blinds', 'script'], 'base_type')
        ], required=False
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance.pk and 'slaves' in self.fields:
            self.fields['slaves'].initial = self.instance.slaves.all()

    def clean_slaves(self):
        if 'slaves' not in self.cleaned_data:
            return
        if not self.cleaned_data['slaves'] or not self.instance:
            return self.cleaned_data['slaves']
        return validate_slaves(self.cleaned_data['slaves'], self.instance)

    def save(self, commit=True):
        obj = super().save(commit=commit)
        if commit and 'slaves' in self.cleaned_data:
            obj.slaves.set(self.cleaned_data['slaves'])
        return obj


class ZwaveKnobComponentConfigForm(BasicZwaveComponentConfigForm):
    min = forms.FloatField(
        initial=0, help_text="Minimum component value."
    )
    max = forms.FloatField(
        initial=100, help_text="Maximum component value."
    )
    zwave_min = forms.FloatField(
        initial=0, help_text="Minimum value expected by Zwave node."
    )
    zwave_max = forms.FloatField(
        initial=99, help_text="Maximum value expected by Zwave node."
    )
    slaves = Select2ModelMultipleChoiceField(
        queryset=Component.objects.filter(
            base_type__in=('dimmer',),
        ),
        url='autocomplete-component',
        forward=(forward.Const(['dimmer', ], 'base_type'),),
        required=False
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance.pk and 'slaves' in self.fields:
            self.fields['slaves'].initial = self.instance.slaves.all()

    def clean_slaves(self):
        if not self.cleaned_data['slaves'] or not self.instance:
            return self.cleaned_data['slaves']
        return validate_slaves(self.cleaned_data['slaves'], self.instance)

    def save(self, commit=True):
        obj = super().save(commit=commit)
        if commit and 'slaves' in self.cleaned_data:
            obj.slaves.set(self.cleaned_data['slaves'])
        return obj

class RGBLightComponentConfigForm(BasicZwaveComponentConfigForm):
    has_white = forms.BooleanField(
        label=_("Has WHITE color channel"), required=False,
    )



import time
from django.utils.translation import gettext_lazy as _
from django.urls import reverse_lazy, path
from django.contrib import admin
from django.contrib import messages
from django.views.generic import TemplateView
from django.shortcuts import render, redirect
from django.utils.safestring import mark_safe
from simo.core.events import GatewayObjectCommand
from simo.core.models import Gateway
from .models import ZwaveNode, NodeValue
from .forms import AdminNodeValueInlineForm, ZwaveGatewaySelectForm




class NodeValueInline(admin.TabularInline):
    form = AdminNodeValueInlineForm
    model = NodeValue
    extra = 0
    fields = (
        'index', 'label',  'type', 'value', 'units',
        'name', 'component_display'
    )
    readonly_fields = (
        'index', 'label', 'units', 'type', 'component_display', 'type'
    )
    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request, obj):
        return False

    def component_display(self, obj):
        if obj.component:
            return mark_safe('<div style="min-width: 200px"><a href="%s"><strong>%s</strong>%s >></a></div>' % (
                obj.component.get_admin_url(),
                obj.component.name,
                ' in %s' % obj.component.zone.name
                if obj.component.zone else ''
            ))
        return ''

    component_display.short_description = _("component")



@admin.register(ZwaveNode)
class ZwaveNodeAdmin(admin.ModelAdmin):
    change_list_template = ''
    list_display = '__str__', 'battery_level', 'alive'
    readonly_fields = (
        'node_id', 'product_name', 'product_type', 'battery_level',
        'stats', 'alive',
    )
    inlines = NodeValueInline,
    actions = [
        'kill_node', 'request_network_update', 'send_node_information',
        'has_node_failed'
    ]

    # def has_delete_permission(self, request, obj=None):
    #     if obj and not obj.alive:
    #         return True
    #     return False

    def response_post_save_change(self, request, obj):
        # Let's give some time for these values to be applied
        # in case zwave node is active and directly reachable
        time.sleep(1)
        obj = ZwaveNode.objects.get(pk=obj.pk)
        return self._response_post_save(request, obj)

    def get_urls(self):
        urls = super().get_urls()
        return [
            path('remove-nodes/', self.remove_nodes)
        ] + urls

    def add_view(self, request, form_url="", extra_context=None):
        ctx = {"opts": self.model._meta, "media": self.media}

        from .gateways import ZwaveGatewayHandler
        zwave_gateways = Gateway.objects.filter(type=ZwaveGatewayHandler.uid)

        if zwave_gateways.count() == 0:
            messages.error(request, "Please create Zwave gateway first!")
            return redirect(reverse_lazy('admin:core_gateway_add'))

        if zwave_gateways.count() == 1:
            request.session['add_nodes_gateway_pk'] = zwave_gateways.first().pk

        if not request.session.get('add_nodes_gateway_pk'):

            ctx['form'] = ZwaveGatewaySelectForm()
            if request.method == 'POST':
                form = ZwaveGatewaySelectForm(request.POST)
                if form.is_valid():
                    request.session['add_nodes_gateway_pk'] = \
                        form.cleaned_data['gateway'].pk
                else:
                    return render(request, 'zwave_node_choose_gateway.html',
                                  ctx)
            else:
                return render(request, 'zwave_node_choose_gateway.html', ctx)


        if request.session.get('add_nodes_gateway_pk'):
            ctx['gateway'] = Gateway.objects.get(
                pk=request.session['add_nodes_gateway_pk']
            )
            if 'new_nodes_only' not in request.GET:
                GatewayObjectCommand(
                    ctx['gateway'], zwave_command='add_node'
                ).publish()
            ctx['gateway'].config['last_controller_command'] = time.time()
            ctx['gateway'].save()
            ctx['zwave_nodes'] = ZwaveNode.objects.filter(
                gateway=ctx['gateway'], is_new=True
            )

            if request.method == 'POST' and 'finish' in request.POST:
                GatewayObjectCommand(
                    ctx['gateway'], zwave_command='cancel_command'
                ).publish()
                request.session.pop('add_nodes_gateway_pk')
                ctx['new_nodes'] = ZwaveNode.objects.filter(
                    gateway=ctx['gateway'], is_new=True
                ).update(is_new=False)
                return redirect(
                    reverse_lazy('admin:zwave_zwavenode_changelist')
                )
        if 'new_nodes_only' in request.GET:
            return render(request, 'zwave_nodes.html', ctx)
        return render(request, 'add_zwave_node.html', ctx )


    def remove_nodes(self, request):
        ctx = {"opts": self.model._meta, "media": self.media}
        from .gateways import ZwaveGatewayHandler

        zwave_gateways = Gateway.objects.filter(type=ZwaveGatewayHandler.uid)

        if zwave_gateways.count() == 0:
            return redirect(reverse_lazy('admin:zwave_zwavenode_changelist'))

        if zwave_gateways.count() == 1:
            request.session['remove_nodes_gateway_pk'] = zwave_gateways.first().pk

        if not request.session.get('remove_nodes_gateway_pk'):

            ctx['form'] = ZwaveGatewaySelectForm()
            if request.method == 'POST':
                form = ZwaveGatewaySelectForm(request.POST)
                if form.is_valid():
                    request.session['remove_nodes_gateway_pk'] = \
                        form.cleaned_data['gateway'].pk
                else:
                    return render(request, 'zwave_node_choose_gateway.html',
                                  ctx)
            else:
                return render(request, 'zwave_node_choose_gateway.html', ctx)

        if request.session.get('remove_nodes_gateway_pk'):
            ctx['gateway'] = Gateway.objects.get(
                pk=request.session['remove_nodes_gateway_pk']
            )
            if 'removed_nodes_only' not in request.GET:
                GatewayObjectCommand(
                    ctx['gateway'], zwave_command='remove_node'
                ).publish()
                request.session['org_nodes_list'] = [
                    str(n) for n in
                    ZwaveNode.objects.filter(gateway=ctx['gateway'])
                ]
            ctx['gateway'].config['last_controller_command'] = time.time()
            ctx['gateway'].save()
            current_nodes = [
                str(n) for n in ZwaveNode.objects.filter(gateway=ctx['gateway'])
            ]
            ctx['zwave_nodes'] = [
                n for n in request.session.get('org_nodes_list')
                if n not in current_nodes
            ]

            if request.method == 'POST' and 'finish' in request.POST:
                GatewayObjectCommand(
                    ctx['gateway'], zwave_command='cancel_command'
                ).publish()
                request.session.pop('remove_nodes_gateway_pk')
                request.session.pop('org_nodes_list')
                return redirect(
                    reverse_lazy('admin:zwave_zwavenode_changelist')
                )
        if 'removed_nodes_only' in request.GET:
            return render(request, 'zwave_nodes.html', ctx)
        return render(request, 'remove_zwave_node.html', ctx )

    def kill_node(self, request, queryset):
        for node in queryset:
            node.kill_node()

    def request_network_update(self, request, queryset):
        for node in queryset:
            node.request_network_update()


    def send_node_information(self, request, queryset):
        for node in queryset:
            node.send_node_information()


    def has_node_failed(self, request, queryset):
        for node in queryset:
            node.has_node_failed()



import os
from random import randint
from dynamic_preferences.preferences import Section
from dynamic_preferences.types import BooleanPreference, StringPreference
from dynamic_preferences.registries import global_preferences_registry
from django.conf import settings


zwave = Section('zwave')


@global_preferences_registry.register
class OZwaveConfigPathCommunity(StringPreference):
    section = zwave
    name = 'ozwave_config_path_community'
    default = os.path.join(settings.VAR_DIR, 'ozwave_config', 'community')
    required = True


@global_preferences_registry.register
class OZwaveConfigPathUser(StringPreference):
    section = zwave
    name = 'ozwave_config_path_user'
    default = os.path.join(settings.VAR_DIR, 'ozwave_config', 'user')
    required = True


@global_preferences_registry.register
class OZwaveNetworkKey(StringPreference):
    section = zwave
    name = 'netkey'
    default = ', '.join([str(hex(randint(0, 255))) for i in range(16)])
    required = True
    help_text = 'DO NOT CHANGE THIS! Unless you know what you are doing!'

